// This file is part of the PSEv3 plugin, released under the BSD 3-Clause License
//
// Andrew Fiore

#include "Helper_Stokes.cuh"

#include <stdio.h>

#ifdef WIN32
#include <cassert>
#else
#include <assert.h>
#endif

//! command to convert floats or doubles to integers
#ifdef SINGLE_PRECISION
#define __scalar2int_rd __float2int_rd
#else
#define __scalar2int_rd __double2int_rd
#endif


/*! \file Helper_Stokes.cu
    	\brief Helper functions required for data handling in Stokes.cu
*/

/*!
	Initialize the total applied force and torque using the net_force
	vector from HOOMD which contains the contributions from external
	and interparticle potentials

	d_net_force		(input)  HOOMD force vector
	d_AppliedForce		(output) Total force experience by the particles
	group_size		(input)  length of vectors
	d_group_members		(input)  index into vectors

*/
__global__ void Stokes_SetForce_kernel(
						Scalar4 *d_net_force,
						float   *d_AppliedForce,
						unsigned int group_size,
						unsigned int *d_group_members
					){

	// Thread idx
	int tidx = blockDim.x * blockIdx.x + threadIdx.x;

	// Do work if thread is in bounds
	if (tidx < group_size) {

		unsigned int idx = d_group_members[ tidx ];
		
		Scalar4 net_force = d_net_force[ idx ];

		d_AppliedForce[ 6*idx     ] = net_force.x;
		d_AppliedForce[ 6*idx + 1 ] = net_force.y;
		d_AppliedForce[ 6*idx + 2 ] = net_force.z;
		d_AppliedForce[ 6*idx + 3 ] = 0.0;
		d_AppliedForce[ 6*idx + 4 ] = 0.0;
		d_AppliedForce[ 6*idx + 5 ] = 0.0;

	}
}

/*!
	Copy velocity computed from solving the hydrodynamic problem
	to the HOOMD velocity array

	d_vel			(output) HOOMD velocity vector
	d_Velocity		(input)  Velocity computed from hydrodynamics
	group_size		(input)  length of vectors
	d_group_members		(input)  index into vectors

*/
__global__ void Stokes_SetVelocity_kernel(
						Scalar4 *d_vel,
						float   *d_Velocity,
						unsigned int group_size,
						unsigned int *d_group_members
						){

	// Thread idx
	int tidx = blockDim.x * blockIdx.x + threadIdx.x;

	// Do work if thread is in bounds
	if (tidx < group_size) {

		unsigned int idx = d_group_members[ tidx ];
		
		Scalar3 vel;
		vel.x = d_Velocity[ 6*idx     ];
		vel.y = d_Velocity[ 6*idx + 1 ];
		vel.z = d_Velocity[ 6*idx + 2 ];

		d_vel[ idx ] = make_scalar4( vel.x, vel.y, vel.z, 0.0 );

	}
}

